/*
This program renders a TrueType font at a given size, collects the
rendered data in an array and generates a function to render a string
using the collected character graymap data.

Its purpose is to embed a font into a program and to render with that
font without ever referring to freetype2 libraries or font files.

In order to compile it, you'll need the freetype2 libraries and compile
as follows:

  gcc -o bfont `pkg-config --cflags --libs freetype2` bfont.c -O3

The O3 flag is useful for generating hash functions.

Command line arguments to the program is as follows:

bfont -f font_file -s size -h hash_func_name -r render_func_name
optional arguments:
   -u    generates the utf8 decoding function also.
   -U    same as -u but the function is declared as static.
   -g    generates a hash function with the given name for the set.
   -G    same as -g, but static function.
   -d dpi
         Default for this value is 96.
You can also use -R instead of -r, which again makes the rendering
function static.

The program reads a list of Unicode character codes to put in the
map. This list is read from the standard input and contains decimal
or hexadecimal character codes. Each line should have either a single
character code or a pair of character codes. The latter case indicates
a range and all characters in the range will be included in the final
character map.

Here is an example input file for the Turkish language character set:

32 126
0xe7
 0x11f
 0x131
 0xf6
 0xfc
 0x15f
 0xc7

0x11e
 0x130
 0xd6
 0xdc
 0x15e
 0x2192
 0xb0

The output, printed on stdout, contains a function with the prototype:

void render_font
        (uint32_t *str, size_t L, 
        int *RW, int *RH, int *RA,
        unsigned char **Rdata);

The name 'render_font' can be changed from the command line. The string 
should be a wide string in Unicode. *RW,*RH and *RA contain width, height
and ascent of the graymap upon return. *Rdata contains one byte per pixel,
in row-major format, allocated using malloc().

This function uses a hash function to convert Unicode characters to glyph indices for its table. This hash function should have the prototype:

  unsigned int fhash(unsigned int V);

It should return a sentinel value if the given character is not in the
map. This sentinel value is the number of characters in the map. The program
can generate this hash function as well, just append the -g switch.

If your input is in utf-8, the program can also add some utf-8 to utf-32 
conversion code to the output. The generated function's prototype is:

int utf8_decode_string
  (uint8_t *input, size_t ilen, uint32_t *output, size_t *olen);

output array should have a size of at least ilen.

So, to sum up, usage is:

bfont -f font_file -s size -h hash_func_name -r render_func_name
optional arguments:
   -u    generates the utf8 decoding function also.
   -U    same as -u but the function is declared as static.
   -g    generates a hash function with the given name for the set.
   -G    same as -g, but static function.
   -d dpi
         Default for this value is 96.
You can also use -R instead of -r, which again makes the rendering
function static.

*/
