#include <ft2build.h>
#include FT_FREETYPE_H
#include <math.h>

#include "font.h"

static FT_Face face;
static  FT_Library ftlib;
static int font_dpi;

static  char *fterrstr(int code) ;

int font_render(wchar_t ch, int *Width, int *Height, int *Ascent, char **Data)
{
   unsigned int gindex;
   int error;

   int width,height,ascent;
   unsigned char *data; 

   FT_GlyphSlot slot;
   FT_Bitmap *ftbitmap; 
   int left_padding;
   /* R stands for rendered */
   int rwidth, rheight, rpitch;
   int i,j;

   gindex= FT_Get_Char_Index(face, ch);
   if (!gindex) 
   {
      fprintf(stderr,"character U%x\n doesn't exist in font.\n", ch);
      return 1;
   }

   error= FT_Load_Glyph(face, gindex, FT_LOAD_RENDER);
   if (error)
   {
      fprintf(stderr, "could not load or render character U%x\n"
                      "freetype error: %s\n", (unsigned)ch, fterrstr(error));
      return 1;
   }

   slot= face->glyph;
   ftbitmap= &slot->bitmap;

   width= slot->metrics.horiAdvance/64;
   height=ftbitmap->rows;
   ascent= slot->bitmap_top;

   left_padding= slot->bitmap_left;

   rwidth= ftbitmap->width;
   rheight= height;
   rpitch= ftbitmap->pitch;

   /* sometimes chars are designed to overflow the given box */
   if (left_padding<0) left_padding=0;

   /* same as above, from right */
   if (width<(rwidth+left_padding)) width= rwidth+left_padding;

   data= calloc(1,width*height);

   for(i=0;i<rheight;i++) 
      for(j=0;j<rwidth;j++)
      {
         int src,dst;
         src= i*rpitch + j;
         dst= i*width + j + left_padding;
         data[dst] = ftbitmap->buffer[src];
      }
	
   *Width= width;
   *Height= height;
   *Ascent= ascent;
   *Data= (char*) data;

   return 0;
}

int font_load(char *file_name,int size,int dpi)
{
   int error;
   static int lib_init= 0;

   if (!lib_init)
   {
      lib_init= 1;
      error= FT_Init_FreeType( &ftlib );
      if (error) 
      {
         fprintf(stderr, "could not initialize freetype library\n"
                         "freetype error: %s\n", fterrstr(error));
         return 1;
      }
   }
                                  // 0 is face index
   error= FT_New_Face(ftlib, file_name, 0, &face);

   if (error)
   {
       fprintf(stderr, "can not load face #0 from file %s\n"
	               "freetype error: %s\n", file_name, fterrstr(error));
       return 1;
   }

	/* FIXME: handle the error properly */
   FT_Set_Char_Size( face, 0, size * 64, 0, dpi);
   font_dpi= dpi;
   return 0; 
}

static  char * fterrstr(int code)
{
   static char buf[200];
#define FT_ERROR_START_LIST     switch(code) {
#define FT_ERRORDEF( e, v, s )  case e: return s;
#define FT_ERROR_END_LIST       }

#include FT_ERRORS_H

   sprintf(buf," Unknown error %d",code);
   return buf;
}

