#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define FILLPERCENT 90
#define COUNT_RESET 10000

#define erintf(...) fprintf (stderr, __VA_ARGS__)

enum { OP_NORM, OP_BOTCUT,  OP_SHIFTXOR, OP_USELESS, OP_SHIFTSUB, };

typedef struct op {
  int code;
  unsigned int arg1,arg2,arg3;
  struct op *link;
} op_t;

typedef struct set {
  op_t *op;
  int len;
  unsigned int *data;
  struct set *next;
} set_t;

op_t *op(int code, int a1,int a2, int a3, op_t *p)
{
  op_t *P;
  P= malloc(sizeof(*P));
  P->code=code; P->arg1= a1; P->arg2= a2; P->arg3= a3; P->link= p;
  return P;
}

set_t *newset(int N)
{
  set_t *T;
  T= malloc(sizeof(*T));
  if (N)
    T->data= malloc(sizeof(int)*N);
  else T->data= NULL;
  T->len= N;
  T->op= NULL;
  T->next= NULL;
  return T;
}

void destroyset(set_t *T) { free(T->data); free(T); }

set_t *copyset(set_t *T)
{
  set_t *S;
  S= newset(T->len);
  memcpy(S->data, T->data, sizeof(int)*T->len);
  S->op= T->op;
  return S;
}

int uincomp(const void *A, const void *B)
{
  const unsigned int a= *(const unsigned int*) A;
  const unsigned int b= *(const unsigned int*) B;
  if (a<b) return -1;
  if (a>b) return 1;
  return 0;
}

void unifyset(set_t *T)
{
  int i,j;
  if (T->len==0) return ;
  qsort(T->data, T->len, sizeof(int), uincomp);
  j= 0;
  for(i=0;i<T->len;i++)
   if (T->data[i]!=T->data[j]) { j++; T->data[j]= T->data[i]; }
  T->len= j+1;
}


set_t *bot_cut(set_t *T)
{
   set_t *C;
   int K,i;

   for(K=T->len-1;K>0;K--) {
     int reqsp,nelts;
     reqsp= T->data[K]-T->data[0]+1;
     nelts= K+1;
     if (nelts*100>=FILLPERCENT*reqsp) break;
   }

   if (2*K<T->len) return NULL;

   C= newset(T->len-K-1);
   // for(i=0;i<C->len;i++) C->data[i]= T->data[i+K+1]- T->data[0];
   // for(i=0;i<C->len;i++) C->data[i]= T->data[i+K+1]- T->data[K];
   for(i=0;i<C->len;i++) C->data[i]= T->data[i+K+1]- T->data[K+1];
   
   C->op= op(OP_BOTCUT, T->data[0], T->data[K], T->data[K+1], T->op);
   return C;
}

set_t *normalize(set_t *T)
{
   int i;
   set_t* C;
   if (T->data[0]==0) return NULL;
   C= newset(T->len);
   for(i=0;i<T->len;i++) C->data[i]= T->data[i]-T->data[0];

   C->op= op(OP_NORM, T->data[0], 0,0, T->op);
   return C;
}

int highest_bit_int(unsigned int V)
{
  int i;
  for(i=31;i>0;i--) if (V & (1<<i)) break;
  return i;
}

int highest_bit(set_t *V)
{
  int i;
  unsigned int mask= 0;
  for(i=0;i<V->len;i++) mask |= V->data[i];

  return highest_bit_int(mask);
}
unsigned int lower[32],upper[32];

unsigned int remove_bit_int(unsigned int A,int B)
{
   unsigned int L,H;
   L= A & lower[B];
   H= A & upper[B];
   return  L | (H>>1);
}


set_t *elim_useless(set_t *V)
{
   int B;
   int HB=highest_bit(V);
   if (HB==0) return NULL;
   for(B=HB;B>=0;B--)
   {
     set_t *Z; int i;
     Z= newset(V->len);
     for(i=0;i<Z->len;i++)
        Z->data[i]= remove_bit_int(V->data[i], B);
     unifyset(Z);
     if (Z->len==V->len)
     {
       Z->op= op(OP_USELESS,B,HB,0,V->op);
       return Z;
     }
     destroyset(Z);
   }
   return NULL;
}


set_t *shift_sub(set_t *V,int shift)
{
   set_t *Z;
   int i;
   Z= newset(V->len);
   for(i=0;i<Z->len;i++) Z->data[i]= V->data[i] - (V->data[i]>>shift);
   unifyset(Z);
   if (Z->len!=V->len) { destroyset(Z); return NULL; }
   Z->op= op(OP_SHIFTSUB, shift, 0,0, V->op);
   return Z;
}

set_t *shift_xor(set_t *T, int shift)
{
  int i;
  set_t *Z;

  Z= newset(T->len);
  for(i=0;i<Z->len;i++) 
      Z->data[i]= T->data[i]^(T->data[i]>>shift);
  unifyset(Z);
  if (Z->len==T->len)
  {
    Z->op= op(OP_SHIFTXOR, shift,0,0,T->op);
    return Z;
  }
  destroyset(Z);
  return NULL;
}

set_t* addq(set_t *L, set_t *V)
{
  if (!V) return L;
  L->next= V;
  return V;
}

void printop(op_t *P)
{
  switch(P->code)
  {
    case OP_NORM: erintf("norm(%d)", P->arg1); break;
    case OP_BOTCUT: erintf("botcut(%d %d %d)", P->arg1, P->arg2,P->arg3); break;
    case OP_SHIFTSUB: erintf("shiftsub(%d)", P->arg1); break;
    case OP_SHIFTXOR: erintf("shiftxor(%d)", P->arg1); break;
    case OP_USELESS: erintf("useless(%d,%d)", P->arg1,P->arg2); break;
  }
}
void printopS(op_t *P)
{
  if (!P) return ;
  printopS(P->link);
  printop(P);
}
void printopR(op_t *P)
{
  if (!P) return ;
  printop(P);
  printopR(P->link);
}


void printset(set_t *T)
{
  int i,j;
  printopS(T->op);
  i=0;
again:
  if (i==T->len) goto done;
  erintf(" %x", T->data[i]);
  i++;
  if (i==T->len) goto done;

  for(j=i;j<T->len;j++)
   if (j-i+1!=T->data[j]-T->data[i-1]) break;

  if (j==i) goto again;

  erintf("-%x", T->data[j-1]);
  i= j;
  goto again;
done:
  erintf("\n");
}

void found(set_t *T)
{
  erintf("Found a solution.\n");
  printset(T);
}

op_t *opreverse(op_t* P,op_t *V)
{
   op_t *R;
   if (P->link) R= opreverse(P->link, P);
   else R= P;
   P->link= V;
   return R;
}


op_t* search(set_t *T)
{
  set_t *first, *last, *R;
  unsigned int count;

  R= bot_cut(T);
  if (R) T= R;

  first= last= T;

  count= 0;

  while(1)
  {
     int i,B;
     T= first;

     if (!T) break;
     if (!T->len) { found(T); return opreverse(T->op,NULL); }

     count++;
     if (count==COUNT_RESET) { erintf("G: "); printset(T); count= 0; }

     last= addq(last, bot_cut(T));

     R= elim_useless(T); 
     if (R) { R->next= first->next; first->next= R; }

     last= addq(last, normalize(T));

     B= highest_bit(T);
     for(i=1;i<=B;i++)
        last= addq(last, shift_sub(T,i));

     B= highest_bit(T);
     for(i=1;i<=B;i++)
        last= addq(last, shift_xor(T,i));

     T= first->next;
     destroyset(first);
     first= T;

  }
  erintf("exhausted?\n");
  exit(1);
  return NULL;
}



void print_bin(unsigned int A)
{
  int i;
  for(i=31;i>=0;i--) if (A&(1<<i)) erintf("1"); else erintf("0");
}

const char *intype(unsigned int A)
{
  int HB;
  HB= highest_bit_int(A);
  if (HB>=16) return "unsigned int";
  if (HB>=8)  return "unsigned short";
  return "unsigned char";
}

void print_htable(set_t *input, op_t *sol)
{
  int S,i;
  unsigned int *pos,*val;
  op_t *P;

  S= 0;
  for(P=sol;P;P=P->link)
    if (P->code==OP_BOTCUT)
      S+= P->arg2-P->arg1+1;
  printf("  static const struct\n  { %s val; %s pos; } htable[%d]= {\n",
    intype(input->data[input->len-1]), intype(input->len+1), S+1);
  pos= malloc(sizeof(int)*(S+1));
  val= malloc(sizeof(int)*(S+1));
  for(i=0;i<S+1;i++) { pos[i]= input->len; val[i]= input->data[0]; }
  for(i=0;i<input->len;i++)
  {
    unsigned int A;
    int p;
    A= input->data[i];
    p= 0;
    for(P=sol;P;)
      switch(P->code)
      {
      case OP_BOTCUT:
           A-= P->arg1;
           if (A<=(P->arg2-P->arg1)) { p+= A; P= NULL; break; }
           A= A - P->arg3 + P->arg1;
           p+= P->arg2-P->arg1+1;
           P= P->link;
           break;
      case OP_NORM:
           A-= P->arg1;
           P= P->link;
           break;
      case OP_SHIFTXOR:
           A ^= A>>P->arg1;
           P= P->link;
           break;
      case OP_SHIFTSUB: 
           A -= A>>P->arg1;
           P= P->link;
           break;
      case OP_USELESS:
           A= remove_bit_int(A, P->arg1);
           P= P->link;
           break;
      }
    if (val[p]!=input->data[0])
    {
      printf("clash between %x and %x?\n", val[p] , input->data[i]);
    }
    val[p]= input->data[i];
    pos[p]= i;
  }
  
  for(i=0;i<S+1;i++) {
      printf(" {0x%x,%d},", val[i], pos[i]);
      if (i&&!(i%6)) printf("\n");
  }
  printf("  };\n");
}

void print_code(op_t* P,int sentinel)
{
  printf("  unsigned int A;\n");
  printf("  int p;\n");
  printf("  A= V;\n");
  printf("  p= 0;\n");
again:
  if (!P) goto done ;

  if (P->code==OP_NORM) goto do_normal;
  if (P->code==OP_USELESS) goto do_useless;
  if (P->code==OP_SHIFTXOR) goto do_shiftxor;
  if (P->code==OP_SHIFTSUB) goto do_shiftsub;
  if (P->code==OP_BOTCUT) goto do_botcut;
next:
  P= P->link;
  goto again;
do_normal:
  printf("  A-= 0x%x;\n", P->arg1);
  goto next;
do_shiftxor:
  printf("  A^= A>>%d;\n", P->arg1);
  goto next;
do_shiftsub:
  printf("  A-= A>>%d;\n", P->arg1);
  goto next;
do_useless:
  if (P->arg2!=P->arg1)
    goto useless_full;
  if (!P->link || P->link->code!=OP_USELESS
               || P->link->arg2!=P->link->arg1)
    printf("  A &= 0x%x;\n", lower[P->arg1]);
  goto next;
useless_full:
  printf("  A= (A&0x%x) | ((A&0x%x)>>1);\n",
            lower[P->arg1], upper[P->arg1]);
  goto next;
do_botcut:
  printf("  A-= 0x%x;\n", P->arg1);
  printf("  if (A<=0x%x) { p+= A; goto done; }\n",
           P->arg2-P->arg1); 
  printf("  A-= 0x%x;\n", P->arg3-P->arg1);
  printf("  p+= %d;\n", P->arg2-P->arg1+1);
  goto next;
done:
  printf("done:\n");
  printf("  if (htable[p].val!=V) return %d;\n", sentinel);
  printf("  return htable[p].pos;\n");
}

#include "obuffer.c"

void prepare_upper_lower() 
{
   int i;
   lower[0]= 0;
   lower[1]= 1;
   for(i=2;i<32;i++)
      lower[i]= lower[i-1] | (lower[i-1]<<1);
   upper[31]= 0;
   upper[30]= (1<<31);
   for(i=29;i>=0;i--)
      upper[i]= upper[i+1] | (upper[i+1]>>1);

#if 0
   for(i=0;i<32;i++)
         { erintf("%2d ", i); print_bin(upper[i]);
           erintf(" "); print_bin(lower[i]); erintf("\n"); }
#endif
}


void print_hash_func(unsigned int *D, int N)
{
   set_t *S,*inp;
   op_t *R,*T;
   int i;

   inp= newset(N);
   for(i=0;i<N;i++) inp->data[i]= D[i];

   prepare_upper_lower();
   T= R= search(copyset(inp));

   erintf("detailed results\n");
   S= copyset(inp);
   while(R)
   {
     switch(R->code)
     {
      case OP_NORM: S= normalize(S); break;
      case OP_BOTCUT: S= bot_cut(S); break; 
      case OP_SHIFTSUB: S= shift_sub(S, R->arg1); break; 
      case OP_SHIFTXOR: S= shift_xor(S,R->arg1); break; 
      case OP_USELESS: S= elim_useless(S); break;
     }
     printset(S);
     S->op= NULL;
     R= R->link;
   }

   printf("{\n");
   print_htable(inp, T);
   print_code(T, inp->len);
   printf("}\n");
}

