#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PS print_spc(spc);

static int senti;


/***********************************************************************
 *                                                                     *
 *                       Binary Search Generator                       *
 *                                                                     *
 ***********************************************************************
 * This program generates a binary search function for a fixed set     *
 * of integers.                                                        *
 *                                                                     *
 * The set is partitioned into two kinds of leaf nodes: ranges and     *
 * single values. Ranges consist of two or more consecutive numbers.   *
 ***********************************************************************/

typedef struct node
{    // 'R'= range
     // 'S'= single
     // 'I'= internal
  int kind;   
  int lo,hi;     
  int cost;
    // left= highest value of the left subtree
    // right= lowest value of the right subtree
    // only valid for internal nodes
  int left_k, right_k;
  struct node *left,*right;
    // for storing sequentially
  struct node *next;
    // this is used for the final values
  int pos;
} node_t;

static int imax2(int a,int b) { return a>b?a:b; }
static int imin2(int a,int b) { return a<b?a:b; }

static node_t *no_new(int kind)
{
  node_t *d;
  d= calloc(1,sizeof(*d));
  d->kind= kind;
  return d;
}

static node_t *no_range(int lo,int hi)
{
  node_t *d;
  d= no_new('R');
  d->lo= lo;
  d->hi= hi;
  d->cost= 2;
  return d;
}

static int rightmost(node_t *d)
{
  switch(d->kind)
  {
  case 'R': return d->hi;
  case 'S': return d->hi;
  case 'I': if (d->right) return rightmost(d->right);
            else if (d->left) return rightmost(d->left);
            else exit(fprintf(stderr,"internal error: (bfont)\n"));
  }
  return 0;
}

static int leftmost(node_t *d)
{
  switch(d->kind)
  {
  case 'R': return d->lo;
  case 'S': return d->lo;
  case 'I': if (d->left) return leftmost(d->left);
            else if (d->right) return leftmost(d->right);
            else exit(fprintf(stderr,"internal error: (bfont)\n"));
  }
  return 0;
}

static node_t *no_internal(node_t *left,node_t *right)
{
  node_t *d;
  d= no_new('I');
  d->left= left;
  d->right= right;
  d->cost= imax2(left->cost,right->cost)+1;
  d->left_k= rightmost(d->left);
  d->right_k= leftmost(d->right);
  return d;
}

static node_t *no_single(int v)
{
  node_t *d;
  d= no_new('S');
  d->cost= 1;
  d->lo= d->hi= v;
  return d;
}

static void print_spc(int x)
{
  int i;
  for(i=0;i<x;i++) printf(" ");
}

static void print_tree(node_t *d)
{
  static int spc= 2;
  if (!d) return ;
  switch(d->kind)
  {
  case 'I': 
     spc+= 5; print_tree(d->left); spc-= 5;
     print_spc(spc);
     printf("I(%d,%d) [%d]\n", d->left_k, d->right_k, d->cost);
     spc+= 5; print_tree(d->right); spc-= 5;
     break;
  case 'R': 
     print_spc(spc);
     printf("R(%d,%d) [%d]\n", d->lo, d->hi, d->cost);
     break;
  case 'S':
     print_spc(spc);
     printf("S(%d) [%d]\n", d->lo, d->cost);
     break;
  }
}

static void print_list(int nnodes,node_t **K)
{
  int i;
  printf("list of nodes:\n");
  for(i=0;i<nnodes;i++)
  {
    print_tree(K[i]);
    printf("----------\n");
  }
}

void print_code(node_t *d,int lo,int hi)
{
  static int spc;
  if (!d) return ;
  switch(d->kind)
  {
  case 'I': 

     if (d->left->kind=='S' && d->right->kind=='S')
     {
       print_spc(spc); printf("if (V==%d) return %d;\n", d->left->lo,d->left->pos);
       print_spc(spc); printf("else if (V==%d) return %d;\n", d->right->lo, d->right->pos);
       // print_spc(spc); printf("else return %d;\n",senti);
       break;
     }

     print_spc(spc); printf("if (V<=%d) {\n",d->left_k);
         spc+= 5; print_code(d->left,lo,d->left_k); spc-= 5;
     print_spc(spc); printf("} else if (V>=%d) {\n", d->right_k);
         spc+= 5; print_code(d->right,d->right_k,hi); spc-= 5;
     print_spc(spc); printf("}\n"); // printf("} else return %d;\n",senti);
     break;
  case 'R': 
     if (hi!=d->hi) { PS printf("if (V>%d) return %d;\n",d->hi,senti); }
     if (lo!=d->lo) { PS printf("if (V<%d) return %d;\n",d->lo,senti);  }
     PS printf("return V%+d;\n",-d->lo+d->pos);
     break;
  case 'S':
     if (d->lo!=lo || d->hi!=hi) 
       { PS printf("if (V!=%d) return %d;\n", d->lo,senti); }
     PS printf("return %d;\n",d->pos);
     break;
  }
}

static node_t *head;

void make_tree(unsigned int *D,int N)
{
  node_t **tgt,*P,*left,*right;
  int i,j,minc,tt,pos;
  senti= N;

  printf("{\n");

  head= NULL;
  tgt= &head;
  for(i=0;i<N;)
  {
    for(j=i+1;j<N;j++)
     if (D[j]-D[j-1]>1) break;
    if (j==i+1)
      P= no_single(D[i]);
    else
      P= no_range(D[i],D[j-1]);
    i= j;
    *tgt= P;
    tgt= &P->next;
  }


  node_t **nodes;
  int nnodes,pa;

  for(nnodes=0,P=head;P;P=P->next) nnodes++;
  nodes= malloc(sizeof(nodes[0])*nnodes);
  for(nnodes=0,P=head;P;P=P->next)
    nodes[nnodes++]= P;

  pos= 0;
  for(i=0;i<nnodes;i++)
  {
    nodes[i]->pos= pos;
    pos+= nodes[i]->hi-nodes[i]->lo+1;
  }

  // print_list(nnodes,nodes);

  while(nnodes>1)
  {
    minc= nodes[0]->cost;
    for(i=1;i<nnodes;i++)
      minc= imin2(minc, nodes[i]->cost);
    // printf("[---minimum cost= %d---]\n", minc);

    for(i=0;i<nnodes;i++)
      if (nodes[i]->cost==minc) break;
    if (i==nnodes) goto fail;

    tt= i;
    pos= tt<(nnodes-1)?2:0;
    pos|= tt>0?1:0;
reswitch:
    switch(pos)
    {
    case 0: goto fail; // impossible! 
    case 1: left= nodes[tt-1]; right= nodes[tt]; pa= tt-1; break; 
    case 2: left= nodes[tt]; right= nodes[tt+1]; pa= tt+1; break;
    case 3: if (nodes[tt-1]->cost < nodes[tt+1]->cost)
                { pos= 1; goto reswitch; }
            else
                { pos= 2; goto reswitch; }
            break;
    }

    nodes[tt]= no_internal(left,right);
    for(i=pa;i<nnodes-1;i++) nodes[i]= nodes[i+1];
    nnodes--;
    // print_list(nnodes,nodes);
  }

  print_code(nodes[0],-1,-1);

  printf("  return %d;\n}\n",senti);

  return ;
fail:
  exit(fprintf(stderr,"internal error\n"));
}


void *__unused_mer_c[]= { (void*) print_list };
